/*
** Copyright (C) 2014 Association of Universities for Research in Astronomy, Inc.
** Contact: mschirme@gemini.edu
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*$Id: gmmps_FoV.c,v 1.3 2011/08/18 22:37:18 gmmps Exp $
 ************************************************************************
 ****      D A O   I N S T R U M E N T A T I O N   G R O U P        *****
 *
 * (c) 2000                         (c) 2000
 * National Research Council        Conseil national de recherches
 * Ottawa, Canada, K1A 0R6          Ottawa, Canada, K1A 0R6
 * All rights reserved              Tous droits reserves
 * 					
 * NRC disclaims any warranties,    Le CNRC denie toute garantie
 * expressed, implied, or statu-    enoncee, implicite ou legale,
 * tory, of any kind with respect   de quelque nature que se soit,
 * to the software, including       concernant le logiciel, y com-
 * without limitation any war-      pris sans restriction toute
 * ranty of merchantability or      garantie de valeur marchande
 * fitness for a particular pur-    ou de pertinence pour un usage
 * pose.  NRC shall not be liable   particulier.  Le CNRC ne
 * in any event for any damages,    pourra en aucun cas etre tenu
 * whether direct or indirect,      responsable de tout dommage,
 * special or general, consequen-   direct ou indirect, particul-
 * tial or incidental, arising      ier ou general, accessoire ou
 * from the use of the software.    fortuit, resultant de l'utili-
 *                                  sation du logiciel.
 *
 ************************************************************************
 *
 * FILENAME
 * gmmps_FoV.c
 *
 * PURPOSE:
 * Extract the exact GMOS field of view
 *
 * FUNCTIONS:
 *  readInData		- Reads in FOV co-ordinates & converts to pixels.
 *
 *# Originially created by:
 *# E.S.O. - VLT project
 *#
 *# Originially created by:
 *# bottini  24/01/00  created
 *
 * SYNOPSIS
 * gmmps_FoV <input> <output> <dataFile> <pixelScale>
 * 
 * DESCRIPTION
 * The Program gmmps_FoV extracts the exact GMOS field of view.
 * The following defines can be modified to keep in account the CCD reading
 * window (only for Y direction) and the CCD offset (only for X direction)
 *
 * The input of the program is the file name, the same used as input for SPOC
 *
 * The output of the program is the same of input but only the objects in the
 * exact GMOS filed of view really visible are stored in.
 *
 * The dataFile contains the cuttoff co-ordinates.
 * All co-ordinates are in arc seconds, and converted to pixels here.
 * Which for example would be:
 * This chops off where:
 *  x<928 || x>5291
 *  y<123 || y>4486 AND
 * This chops off the 4 corners. 4 Co-ord's are:
 * TopLeft     A( 928,4296) to B(1118,4486):slope,Ay-By/Ax-Bx, xo=By-slope*Bx
 * TopRight    C(5101,4486) to D(5291,4296):slope,(Cy-Dy)/(Cx-Dx),xo=Dy-slope*Dx
 * BottomLeft  E( 928, 313) to F(1118, 123):slope,(Ey-Fy)/(Ex-Fx),xo=Fy-slope*Fx
 * BottomRight G(5101, 123) to H(5291, 313):slope,(Gy-Hy)/(Gx-Hx),xo=Hy-slope*Hx
 *
 *          XCor2         XCor1
 *           |              |
 *              B        C
 * YCor2-     /           \
 *          A               D
 *  
 *          E               H
 * YCor1-     \           /
 *              F        G
 * 
 *
 * Arguments:
 *   <input> full input file name.  File format:
 * 	ID, RA, DEC, X, Y, posX, posY, dX, dY, tilt, mag, priority, type
 *
 *   <output> full output file name.
 * 	ID, RA, DEC, X, Y, posX, posY, dX, dY, tilt, mag, priority, type
 *
 *   <pixelScale> 
 *
 *   <data> full name of file containing cutoff co-ordinates.  All values are
 *   in arc seconds, and converted here.  File format:
 *   keyword 	value
 * 
 *   Please note!!!!: The keyword is predefined and can be in any order in the
 *   file, BUT, must contain all of those keywords:
 *   XEND, YEND, XSTART, YSTART, XCORNER1, XCORNER2, YCORNER1, YCORNER2, YDELTA
 *
 *INDENT-OFF*
 * $Log: gmmps_FoV.c,v $
 * Revision 1.3  2011/08/18 22:37:18  gmmps
 * Bugfixes and small improvements, focus on mask design algorithm
 *
 * Revision 1.2  2011/04/25 18:27:33  gmmps
 * Forked from 0.401.12 .
 *
 * Revision 1.1  2011/01/24 20:02:14  gmmps
 * Compiled for RedHat 5.5 32 and 64 bit.
 *
 * Revision 1.2  2002/08/21 01:59:49  callen
 * this is the fix to the bug where declinations from -1hr to 0 would become
 * positive.  Makefile was fixed to link in copysign (problem was a negative
 * zero and trying to get the sign by h/abs(h) where h was the hour component).
 * Actual fix was in gmmps_FoV.c, which exists to cull object not in the field of
 * view.
 *
 * Revision 1.1.1.1  2002/07/19 00:02:09  callen
 * importing gmmps as recieved from Jennifer Dunn
 * gmmps is a skycat plugin and processes for creating masks
 *
 * Revision 1.11  2001/11/27 23:05:39  dunn
 * Change of name and dropping vltPort.h
 *
 * Revision 1.10  2001/11/01 23:18:15  dunn
 * Changes so that all val's inthe fov.dat file are in
 * arc seconds, rather than pixels.
 *
 * Revision 1.9  2001/10/11 21:06:05  dunn
 * Change in co-ordinates & added pixel scale as input.
 *
 * Revision 1.8  2001/08/27 22:18:34  dunn
 * Fixed pos_x/y subtraction from xccd
 *
 * Revision 1.7  2001/08/22 22:55:24  dunn
 * *** empty log message ***
 *
 * Revision 1.6  2001/08/13 21:22:10  dunn
 * Added more informational messages.
 *
 * Revision 1.5  2001/08/02 20:08:19  dunn
 * Added debug statements.
 *
 * Revision 1.4  2001/07/17 19:45:51  dunn
 * Changed it to read in parameters from a file, fov.dat
 *
 * Revision 1.3  2001/04/25 20:39:26  dunn
 * Changed comments.
 *
 * Revision 1.2  2001/04/25 17:01:01  dunn
 * Initial revisions.
 *
 *INDENT-ON*
 *
 ****      D A O   I N S T R U M E N T A T I O N   G R O U P        *****
 ************************************************************************
*/

/*
 *  Local Defines
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

/*
 *  Local Types
 */

typedef struct {
  float	start;		/* Starting co-ordinate.*/
  float	end;		/* End co-ordinate.	*/
  float	cor1;		/* First corner.	*/
  float	cor2;		/* Second corner.	*/
  float delta;		/* Delta, slope scaling factor.	*/
} GM_COORDS;

static  GM_COORDS XVal;		/* Contain X co-ordinate values.	*/
static  GM_COORDS YVal;		/* Contain Y co-ordinate values.	*/


/*
 *  Function Prototypes *** ALL LOCAL FUNCTIONS TO BE PROTOTYPED ***
 */

int readInData ( char *, float );


/*
 *  Data Structures
 */


/*
 *  Macros
 */

/*
 ************************************************************************
 *+
 * FUNCTION: main
 *
 * RETURNS: int, [short description]
 *
 * DESCRIPTION: Main entry
 *
 * [NOTES:]:
 *-
 ************************************************************************
 */

int main (int argc, char *argv[]) {
  int num;
  int numDropped, numGuides;
  int numP1, numP2, numP3;		/* Num objects dropped.	*/
  float X,Y,dX,dY, xoff;
  float pX, pY;
  float Ax,Ay,Bx,By,Cx,Cy,Dx,Dy,m1,q1,m2,q2;
  float Ex,Ey,Fx,Fy,Gx,Gy,Hx,Hy,m3,q3,m4,q4;
  double rah,ram,ras;
  double decg,decm,decs;
  double ra,dec;
  float tilt;		/* Slit tilt angle.			*/
  float mag;		/* Magniture.				*/
  char priority, type;	/* Priority of object & slit type.	*/
  char inputname[200],outname[200], dataname[200];
  FILE *inputfile, *outfile;		/* File ptrs.			*/
  int		cntr;			/* Num of var. read from file.	*/
  char	oneLine[256];		/* One line of input.		*/
  float	pixelScale;		/* Pixel scale.			*/
  
  /*
   * COMMAND LINE INPUT 
   */
  
  if(argc==5) {
    strcpy(inputname,argv[1]);
    strcpy(outname,argv[2]);
    strcpy(dataname,argv[3]);
    pixelScale = atof(argv[4]);
  }
  else {
    printf("gmmps_FoV: WRONG INPUT COMMAND LINE: EXIT\n");
    printf("Usage: gmmps_FoV <inputFile> <output> <data> <pixelScale>.\n");
    printf("       <inputFile> full input file name.\n");
    printf("       <output> full output file name.\n");
    printf("       <data> name of file containing cut-off coordinates.\n");
    printf("       <pixelScale> conv. factor to get pixels from arcs.\n");
    return -1;
  }
  
  /*
   *  Read in the data file and get the co-ordinates.
   */
  
  if ( readInData( dataname, pixelScale ) != 0 ) {
    printf("gmmps_FoV: Bad data file: %s.\n", dataname);
    return -1;
  }
  
  /*
   * Initialize slopes and x0's.
   */
  
  m1 = m2 = m3 = m4 = 0;
  q1 = q2 = q3 = q4 = 0;
  
  /*
   *  WARNING, you will see this xoff value used, but since
   *  it is zero it nulls out some values.  I have left it in just
   *  in case someone wants to change the corners a bit.
   */
  
  xoff=0;
  
  /*
   *  Calculate the corner slopes.
   *  The Top Left Corner.
   */
  
  Ax = XVal.start;
  Ay = YVal.end   - YVal.cor2;
  Bx = XVal.start + XVal.cor2;
  By = YVal.end;
  m1 = (Ay - By) / (Ax-Bx);
  q1 = (By - m1 * Bx) + m1 * xoff;
  
  /*
   *  Calculate the corner slopes.
   *  The Top Right Corner.
   */
  
  Cx = XVal.end - XVal.cor1;
  Cy = YVal.end;
  Dx = XVal.end;
  Dy = YVal.end-YVal.cor2;
  m2 = (Cy - Dy) / (Cx - Dx);
  q2 = (Dy - m2 * Dx) + m2 * xoff;
  
  /*
   *  Calculate the corner slopes.
   *  The Bottom Left Corner.
   */
  
  Ex = XVal.start;
  Ey = YVal.start + YVal.cor1;
  Fx = XVal.start+XVal.cor2;
  Fy = YVal.start;
  m3 = (Ey - Fy) / (Ex - Fx);
  q3 = (Fy - m3 * Fx) - YVal.delta + m3 * xoff;
  
  /*
   *  Calculate the corner slopes.
   *  The Bottom Right Corner.
   */
  
  Gx = XVal.end - XVal.cor1;
  Gy = YVal.start;
  Hx = XVal.end;
  Hy = YVal.start + YVal.cor1;
  m4 = (Gy - Hy) / (Gx - Hx);
  q4 = (Hy - m4 * Hx) - YVal.delta + m4 * xoff;
  
  /*printf("m1=%f, q1=%f, m2=%f, q2=%f.\n", m1, q1, m2, q2 );*/
  
  /*
   *  Open files for reading and writing.
   */
  
  inputfile  = fopen(inputname, "r");
  outfile    = fopen(outname, "w");
  numDropped = 0;
  numGuides  = numP1 = numP2 = numP3 = 0;
  
  /*
   *  While there are lines in the input file, read in :
   *  num, ra, dec, x, y, d, priority and type.
   */
  
  while ( fgets( oneLine, 256, inputfile) != NULL ) {
    /*
     *  Ignore lines beginning with comments, or less than 2 chars
     *  or that are too long and don't finish with a carriage return.
     */
    
    if ( oneLine[0] == '#' || strlen(oneLine) < 2 ) continue;
    if ( oneLine[ strlen(oneLine) - 1] != '\n' ) continue;
    
    
    /*
     *  Scan the line for the correct format.  If incorrect then don't
     *  continue.
     */
    /*printf("oneline=%s\n", oneLine);*/
    
    if ( (cntr = sscanf(oneLine,"%d %lf:%lf:%lf %lf:%lf:%lf %f %f %f %f %f %f %f %f %c %c",
			&num,&rah,&ram,&ras,&decg,&decm,&decs,&X,&Y,&pX,&pY,&dX,&dY,&tilt,
			&mag, &priority, &type)) ==EOF || cntr != 17 ) {
      printf("            PROGRAM ABNORMALY TERMINATED\n");
      printf("            Bad inputline: %s <%d>\n", oneLine, cntr );
      printf("----------------------------------------------------\n"); 
      (void ) fclose(inputfile);
      (void ) fclose(outfile);
      return -1;
    }
      
    /*printf("Reading in data in gmmpsFoV.c")
      printf("%d %f:%f:%f %f:%f:%f %f %f %f %f %f %c %c\n",num,rah,ram,ras,decg,decm,decs,X,Y,dX,dY,tilt, priority, type); */
      
      
    /*
     *  Convert ra/dec to decimal point degrees.
     */
    
    if ( rah != 0.0 )
      ra=(rah/abs(rah))*(abs(rah)*15+ram/4+ras/240);
    else
      ra=(rah)*15+ram/4+ras/240;
    
    /* This is the sign error, it fails for negative zero
       doesn't work because negative zero is ignored, the answer is to use
       copysign(x,y) which takes the sign of y and applies it to the abs 
       value of x
       
       if ( decg != 0.0 )
       dec=(decg/abs(decg))*(abs(decg)+decm/60+decs/3600);
       else
       dec=(decg)+decm/60+decs/3600;
    */
      
    /* new version of the above @@cba */
    dec = copysign(1.0,-20.0);
    dec = copysign( (double)1.0, (double)decg)*(abs(decg)+decm/60+decs/3600);
    
    
    /*printf("%d %f %f %f %f %f %f %f %c %c\n",
      num,ra,dec,X,Y,dX,dY,tilt, priority, type);*/
    
    /*
     *  Calculate the real slit position, x_ccd+(slitpos_x/pixelScale), 
     *  and do the same for y.
     */
    
    X += ( pX / pixelScale ); 
    Y += ( pY / pixelScale ); 
    
    
    /*
     *  If this item falls in one of the corners,
     *  then drop the item, otherwise put to output file.
     */
    if(Y <= m1 * X + q1 && Y <= m2 * X + q2 && Y >= YVal.start && 
       Y <= YVal.end - YVal.delta && X >= XVal.start && X <= XVal.end) {
	  
      if(Y>=m3*X+q3 && Y>=m4*X+q4 && Y>=1 ) {
	fprintf (outfile,"%d %f %f %f %f %f %f %f %f %f %f %c %c\n",
		 num, ra, dec, (X-(pX/pixelScale)), (Y-(pY/pixelScale)),
		 pX, pY, dX, dY, tilt, mag, priority, type);
      }
      else {
	numDropped++;
	if      ( priority == '0' ) numGuides++;
	else if ( priority == '1' ) numP1++;
	else if ( priority == '2' ) numP2++;
	else if ( priority == '3' ) numP3++;
      }
    }
    else {
      numDropped++;
      if      ( priority == '0' ) numGuides++;
      else if ( priority == '1' ) numP1++;
      else if ( priority == '2' ) numP2++;
      else if ( priority == '3' ) numP3++;
    }
  }

  
  /*
   *  Close files.
   */
  
  (void ) fclose(inputfile);
  (void ) fclose(outfile);
  
  printf("Field of View Calculations:\n");
  printf("   Num objects out of view: %d.\n", numDropped );
  printf("   Dropped Object Breakdown:\n" );
  printf("   Acquisition objects:   %d\n",numGuides );
  printf("   Priority 1 :     %d\n",numP1 );
  printf("   Priority 2 :     %d\n",numP2 );
  printf("   Priority 3 :     %d\n",numP3 );
  printf("   Ignored objects:   %d\n",(numDropped - (numGuides+numP1+numP2+numP3) ) );
  printf("----------------------------------------------------\n"); 
  return 0;
}

/*
************************************************************************
*+
* FUNCTION: readInData
*
* RETURNS: int, [0=success ]
*
* DESCRIPTION: 
* Read in the dimensions for the corners.
*
* [NOTES:]:
*-
************************************************************************
*/

int readInData (char *fileName,	float pixelScale) {
  FILE	*filePtr;	/* File Ptr.			*/
  char	oneLine[256];	/* One line of input.		*/
  char	keyword[40];	/* Keyword.			*/
  int	nLines;		/* Number of lines read in.	*/
  int	num;		/* Num of var. read from file.	*/
  float	value;		/* Value read in.		*/
  GM_COORDS *ptr;       /* Ptr to co-ordinates.		*/
  
  if ( fileName[0] == '\0' || ( filePtr = fopen(fileName, "r")) == NULL ) {
    printf("Failed to open %s for reading cut-off co-ordinates.\n", fileName );
    return( -1 );
  }


  /*
   *  Read in the file, saving the values.
   *  Warning, not checking for multiple lines of the same keyword.
   */
  
  nLines=0;
  while ( fgets( oneLine, 256, filePtr) != NULL ) {
    /*
     *  Ignore lines beginning with comments, or less than 2 chars
     *  or that are too long and don't finish with a carriage return.
     */
    
    if ( oneLine[0] == '#' || strlen(oneLine) < 2 ) continue;
    if ( oneLine[ strlen(oneLine) - 1] != '\n' ) continue;
    
    /*
     *  Scan the line for the correct format.  If incorrect then don't
     *  continue.
     */
    
    if ( (num = sscanf( oneLine,"%s %f", keyword, &value ) ) == EOF && num != 2) {
      printf("Invalid num of items in the line: %s.\n", oneLine );
      fclose( filePtr );
      return( -1 );
    }
    
    if ( keyword[0] == 'X' ) {
      ptr = &XVal;
    } 
    else if ( keyword[0] == 'Y' ) {
      ptr = &YVal;
    }
    else {
      printf("Invalid keyword: %s, skipping.\n", keyword );
      continue;
    }
    
    if ( keyword[1] == 'E' ) {
      /*
       *  Have END keyword, get & save the value.
       */
      ptr->end = value/pixelScale;
    }
    else if ( keyword[1] == 'S' ) {
      /*
       *  Have START keyword, get & save the value.
       */
      ptr->start = value/pixelScale;
    }
    else if ( keyword[1] == 'D' && keyword[0] == 'Y' && keyword[2] == 'E' ) {
      /*
       *  Have DELTA keyword, get & save the value.
       */
      ptr->delta = value/pixelScale;
    }
    else if ( strlen(keyword) != 8 ) {
      continue;
    }
    else if ( keyword[7] == '1' ) {
      /*
       *  Have CORNER1 keyword, get & save the value.
       */
      ptr->cor1 = value/pixelScale;
    }
    else if ( keyword[7] == '2') {
      /*
       *  Have CORNER2 keyword, get & save the value.
       */
      ptr->cor2 = value/pixelScale;
    }
    else
      printf("Invalid keyword: %s, skipping it.\n", keyword );

    nLines++;
  }/* End while reading in lines... */

  if ( nLines !=  9 ) {
    printf("Not all keywords read...... only have %d.\n", nLines);
    fclose( filePtr );
    return (-1);
  }
  
  (void) fclose( filePtr );
  return (0);
}
